from rply import LexerGenerator
import re


def strip_comments(s: str):
    ssub = re.sub(r"--.*$", "", s, flags=re.RegexFlag.MULTILINE)
    return ssub


TOKENS = {
    "PROGRAM": r"\bprogram\b",
    "LAMBDA": r"\blam\b",
    "FORCE": r"\bforce\b",
    "DELAY": r"\bdelay\b",
    "BUILTIN": r"\bbuiltin\b",
    "CON": r"\bcon\b",
    "ERROR": r"\berror\b",
    "PAREN_OPEN": r"\(",
    "PAREN_CLOSE": r"\)",
    "BRACK_OPEN": r"\[",
    "BRACK_CLOSE": r"\]",
    "CARET_OPEN": r"\<",
    "CARET_CLOSE": r"\>",
    # there may be an escaped " inside the string (marked by an uneven number of \ before it)
    # the " at the end must be preceded by an even number of \ -- it is not escaped
    "TEXT": r'"(([^\n\r"]|(?<!\\)(\\\\)*\\")*(?<!\\)(\\\\)*)"',
    "COMMA": r",",
    "DOT": r"\.",
    "NUMBER": r"[-\+]?\d+",
    "BOOL": r"\b(True|False)\b",
    "I": r"\bI\b",
    "B": r"\bB\b",
    "LIST": r"\bList\b",
    "MAP": r"\bMap\b",
    "CONSTR": r"\bConstr\b",
    "NAME_NON_SPECIAL": r"[\w_~'][\w\d_~'!#]*",
    "HEX": r"#([\dabcdefABCDEF][\dabcdefABCDEF])*",
}


class Lexer:
    def __init__(self):
        self.lexer = LexerGenerator()

    def _add_tokens(self):
        for k, v in TOKENS.items():
            self.lexer.add(k, v)
        self.lexer.ignore("\s+")

    def get_lexer(self):
        self._add_tokens()
        return self.lexer.build()



from ast import NodeVisitor
from .ast import *


def _int_to_bits(x: int):
    return bin(x)[2:]


class BitWriter:
    """
    BitWriter turns a string of '0's and '1's into a list of bytes.
    Finalization pads the bits using '0*1' if not yet aligned with the byte boundary.
    """

    def __init__(self):
        self._parts = []
        self._n = 0

    @property
    def length(self):
        """Number of bits written so far"""
        return self._n

    def write(self, bit_chars):
        """
        Write a string of '0's and '1's to the BitWriter.
        :param bit_chars: string
        """
        assert all(
            c in ("0", "1") for c in bit_chars
        ), "did pass invalid value to write in bitwriter"

        self._parts.append(bit_chars)
        self._n += len(bit_chars)

    def write_fixed_width_int(self, i: int, width: int):
        """
        Write an integer with fixed width (number of bits)
        :param i: int
        :param width: int
        """
        self.write(self.pad_zeroes(_int_to_bits(i), width))

    def write_nibble(self, nibble: int):
        """
        Write a nibble to the BitWriter.
        :param nibble: int
        """
        self.write_fixed_width_int(nibble, 4)

    def write_byte(self, byte: int):
        """
        Write a byte to the BitWriter.
        :param byte: int
        """
        self.write_fixed_width_int(byte, 8)

    def write_bytes(self, byts: bytes):
        """
        Write a bytestring to the BitWriter.
        :param bytes: bytearray
        """
        self.pad_to_byte_boundary(True)
        n = len(byts)
        pos = 0
        while pos < n:
            n_chunk = min(n - pos, 255)
            self.write_byte(n_chunk)
            for byte in byts[pos : pos + n_chunk]:
                self.write_byte(byte)
            pos += n_chunk
        self.write_byte(0)

    @staticmethod
    def pad_zeroes(s, n):
        return s.zfill(n)

    def pad_to_byte_boundary(self, force=False):
        """
        Add padding to the BitWriter in order to align with the byte boundary.
        If 'force == True' then 8 bits are added if the BitWriter is already aligned.
        :param force: bool
        """
        n_pad = 0
        if self._n % 8 != 0:
            n_pad = 8 - self._n % 8
        elif force:
            n_pad = 8

        if n_pad != 0:
            padding = ["0"] * n_pad
            padding[n_pad - 1] = "1"

            self._parts.append("".join(padding))
            self._n += n_pad

    def finalize(self, force=True):
        """
        Pads the BitWriter to align with the byte boundary and returns the resulting bytes.
        :param force: bool - force padding (will add one byte if already aligned)
        :return: list[int]
        """
        self.pad_to_byte_boundary(force)

        chars = "".join(self._parts)

        bytes_list = []

        for i in range(0, len(chars), 8):
            byte_chars = chars[i : i + 8]
            byte = int(byte_chars, 2)

            bytes_list.append(byte)

        return bytes(bytes_list)

    def write_int(self, i: int, signed: bool):
        i = int(i)
        assert signed or i >= 0, f"Tried to encode unsigned int {i} but is negative"
        zigzagged = zigzag(i, signed)
        bitstring = pad_zeroes(_int_to_bits(zigzagged), 7)

        # split every 7th
        parts = list(chunkstring(bitstring, 7))
        parts.reverse()

        # write all but the last
        for chunk in parts[:-1]:
            self.write("1" + chunk)
        # write the last
        self.write("0" + parts[-1])


def zigzag(i: int, signed: bool):
    """Zigzag-encode an integer"""
    if not signed:
        return i
    else:
        if i < 0:
            return 2 * (-i) - 1
        else:
            return 2 * i


def chunkstring(string, length):
    """Chunk a string into parts of fixed length"""
    return (string[0 + i : length + i] for i in range(0, len(string), length))


def pad_zeroes(bits, n):
    """Prepends zeroes to a bit-string so that 'len(result) == n'."""
    if len(bits) % n != 0:
        n_pad = n - (len(bits) % n)
        bits = n_pad * "0" + bits
    return bits


class FlatEncodingVisitor(NodeVisitor):
    def __init__(self, bw: typing.Optional[BitWriter] = None):
        self.bit_writer = BitWriter() if bw is None else bw

    def visit_Program(self, n: Program):
        for i in n.version:
            self.bit_writer.write_int(i, signed=False)
        self.visit(n.term)

    def visit_Variable(self, n: Variable):
        self.bit_writer.write("0000")
        # this requires the deBrujin encoding rather than the actual variable name
        self.bit_writer.write_int(int(n.name), signed=False)

    def visit_Delay(self, n: Delay):
        self.bit_writer.write("0001")
        self.visit(n.term)

    def visit_Lambda(self, n: Lambda):
        self.bit_writer.write("0010")
        self.visit(n.term)

    def visit_Apply(self, n: Apply):
        self.bit_writer.write("0011")
        self.visit(n.f)
        self.visit(n.x)

    def visit_Constant(self, n: Constant):
        self.bit_writer.write("0100")
        self.bit_writer.write("1")
        ConstantTypeFlatEncodingVisitor(self.bit_writer).visit(n)
        self.bit_writer.write("0")
        ConstantValueFlatEncodingVisitor(self.bit_writer).visit(n)

    def visit_Force(self, n: Force):
        self.bit_writer.write("0101")
        self.visit(n.term)

    def visit_Error(self, n: Error):
        self.bit_writer.write("0110")

    def visit_BuiltIn(self, n: BuiltIn):
        self.bit_writer.write("0111")
        # write index of uplc builtin
        index = n.builtin.value
        self.bit_writer.write_fixed_width_int(index, width=7)

    def visit_BuiltinUnit(self, n: BuiltinUnit):
        self.visit_Constant(n)

    def visit_BuiltinBool(self, n: BuiltinBool):
        self.visit_Constant(n)

    def visit_BuiltinInteger(self, n: BuiltinInteger):
        self.visit_Constant(n)

    def visit_BuiltinByteString(self, n: BuiltinByteString):
        self.visit_Constant(n)

    def visit_BuiltinString(self, n: BuiltinString):
        self.visit_Constant(n)

    def visit_BuiltinPair(self, n: BuiltinPair):
        self.visit_Constant(n)

    def visit_BuiltinList(self, n: BuiltinList):
        self.visit_Constant(n)

    def visit_PlutusData(self, n: PlutusData):
        self.visit_Constant(n)

    def visit_PlutusInteger(self, n: PlutusInteger):
        self.visit_Constant(n)

    def visit_PlutusByteString(self, n: PlutusByteString):
        self.visit_Constant(n)

    def visit_PlutusList(self, n: PlutusList):
        self.visit_Constant(n)

    def visit_PlutusMap(self, n: PlutusMap):
        self.visit_Constant(n)

    def visit_PlutusConstr(self, n: PlutusConstr):
        self.visit_Constant(n)


class ConstantValueFlatEncodingVisitor(NodeVisitor):
    def __init__(self, bw: typing.Optional[BitWriter] = None):
        self.bit_writer = BitWriter() if bw is None else bw

    def visit_BuiltinUnit(self, n: BuiltinUnit):
        pass

    def visit_BuiltinBool(self, n: BuiltinBool):
        self.bit_writer.write(str(int(n.value)))

    def visit_BuiltinInteger(self, n: BuiltinInteger):
        self.bit_writer.write_int(n.value, signed=True)

    def visit_BuiltinByteString(self, n: BuiltinByteString):
        self.bit_writer.write_bytes(n.value)

    def visit_BuiltinString(self, n: BuiltinString):
        self.bit_writer.write_bytes(n.value.encode("utf8"))

    def visit_BuiltinPair(self, n: BuiltinPair):
        self.visit(n.l_value)
        self.visit(n.r_value)

    def visit_BuiltinList(self, n: BuiltinList):
        for v in n.values:
            self.bit_writer.write("1")
            self.visit(v)
        self.bit_writer.write("0")

    def visit_PlutusData(self, n: PlutusData):
        self.bit_writer.write_bytes(plutus_cbor_dumps(n))

    def visit_PlutusInteger(self, n: PlutusInteger):
        self.visit_PlutusData(n)

    def visit_PlutusByteString(self, n: PlutusByteString):
        self.visit_PlutusData(n)

    def visit_PlutusList(self, n: PlutusList):
        self.visit_PlutusData(n)

    def visit_PlutusMap(self, n: PlutusMap):
        self.visit_PlutusData(n)

    def visit_PlutusConstr(self, n: PlutusConstr):
        self.visit_PlutusData(n)


class ConstantTypeFlatEncodingVisitor(NodeVisitor):
    def __init__(self, bw: typing.Optional[BitWriter] = None):
        self.bit_writer = BitWriter() if bw is None else bw

    def visit_BuiltinUnit(self, n: BuiltinUnit):
        self.bit_writer.write_nibble(3)

    def visit_BuiltinBool(self, n: BuiltinBool):
        self.bit_writer.write_nibble(4)

    def visit_BuiltinInteger(self, n: BuiltinInteger):
        self.bit_writer.write_nibble(0)

    def visit_BuiltinByteString(self, n: BuiltinByteString):
        self.bit_writer.write_nibble(1)

    def visit_BuiltinString(self, n: BuiltinString):
        self.bit_writer.write_nibble(2)

    def visit_BuiltinPair(self, n: BuiltinPair):
        self.bit_writer.write_nibble(7)
        self.bit_writer.write("1")
        self.bit_writer.write_nibble(7)
        self.bit_writer.write("1")
        self.bit_writer.write_nibble(6)
        self.bit_writer.write("1")
        self.visit(n.l_value)
        self.bit_writer.write("1")
        self.visit(n.r_value)

    def visit_BuiltinList(self, n: BuiltinList):
        self.bit_writer.write_nibble(7)
        self.bit_writer.write("1")
        self.bit_writer.write_nibble(5)
        self.bit_writer.write("1")
        self.visit(n.sample_value)

    def visit_PlutusData(self, n: PlutusData):
        self.bit_writer.write_nibble(8)

    def visit_PlutusInteger(self, n: PlutusInteger):
        self.visit_PlutusData(n)

    def visit_PlutusByteString(self, n: PlutusByteString):
        self.visit_PlutusData(n)

    def visit_PlutusList(self, n: PlutusList):
        self.visit_PlutusData(n)

    def visit_PlutusMap(self, n: PlutusMap):
        self.visit_PlutusData(n)

    def visit_PlutusConstr(self, n: PlutusConstr):
        self.visit_PlutusData(n)



from ast import NodeVisitor
from typing import Callable

from .ast import *

UPLC_TAG_WIDTHS = {
    "term": 4,
    "type": 3,
    "constType": 4,
    "builtin": 7,
    "constant": 4,
    "kind": 1,
}


def parse_raw_byte(b):
    """
    Parses a single byte in the Plutus-core byte-list representation of an int
    :param b: int
    :return: int
    """
    return b & 0b01111111


def raw_byte_is_last(b):
    """
    Returns true if 'b' is the last byte in the Plutus-core byte-list representation of an int.
    :param b: int
    :return: bool
    """
    return (b & 0b10000000) == 0


def bytes_to_int(bytes):
    """
    Combines a list of Plutus-core bytes into a int (leading bit of each byte is ignored).
    Differs from int.from_bytes because only 7 bits are used from each byte.
    :param bytes: list[int]
    :return: int
    """
    value = 0

    n = len(bytes)

    for i in range(n):
        b = bytes[i]

        # 7 (not 8), because leading bit isn't used here
        value += b * (2 ** (i * 7))

    return value


def unzigzag(value: int, signed: bool):
    """
    Unapplies zigzag encoding
    :return: int
    """
    if not signed:
        return value
    else:
        if value % 2 == 0:
            return value // 2
        else:
            return -((value + 1) // 2)


class UplcDeserializer:
    def __init__(self, bits: str):
        self._bits = bits
        self._pos = 0

    def tag_width(self, category: str) -> int:
        assert category in UPLC_TAG_WIDTHS, f"unknown tag category {category}"

        return UPLC_TAG_WIDTHS[category]

    def built_in_fun(self, id: int) -> BuiltInFun:
        return BuiltInFun(id)

    def read_linked_fixed_width_integer_list(self, elem_size: int) -> List[int]:
        nil_or_cons = self.read_bit()

        if nil_or_cons == 0:
            return []
        else:
            elem = self.read_fixed_width_integer(elem_size)
            return [elem] + self.read_linked_fixed_width_integer_list(elem_size)

    def read_term(self) -> AST:
        tag = self.read_tag("term")

        if tag == 0:
            return self.read_variable()
        elif tag == 1:
            return self.read_delay()
        elif tag == 2:
            return self.read_lambda()
        elif tag == 3:
            return self.read_apply()
        elif tag == 4:
            return self.read_constant()
        elif tag == 5:
            return self.read_force()
        elif tag == 6:
            return Error()
        elif tag == 7:
            return self.read_builtin()
        else:
            raise ValueError(f"term tag {tag} unhandled")

    def read_integer(self, signed: bool = False) -> int:
        byts = []

        b = self.read_byte()
        byts.append(b)

        while not raw_byte_is_last(b):
            b = self.read_byte()
            byts.append(b)

        res = bytes_to_int([parse_raw_byte(b) for b in byts])

        res = unzigzag(res, signed)

        return res

    def move_to_byte_boundary(self, force=False):
        """
        Moves position to the next byte boundary.

        Args:
            force (bool): If True, move to the next byte boundary even if already at one.

        Returns:
            None
        """
        if self._pos % 8 != 0:
            n = 8 - self._pos % 8
            self.read_bits(n)
        elif force:
            self.read_bits(8)

    def read_bytes(self) -> bytes:
        self.move_to_byte_boundary(True)

        byts = []

        n_chunk = self.read_byte()

        while n_chunk > 0:
            for _ in range(n_chunk):
                byts.append(self.read_byte())

            n_chunk = self.read_byte()

        return bytes(byts)

    def read_builtin_byte_string(self) -> BuiltinByteString:
        byts = self.read_bytes()

        return BuiltinByteString(byts)

    def read_builtin_string(self) -> BuiltinString:
        byts = self.read_bytes()

        s = byts.decode("utf8")

        return BuiltinString(s)

    def read_list(self, typed_reader: Callable[[], Constant]) -> List[Constant]:
        items = []

        while self.read_bit() == 1:
            items.append(typed_reader())

        return items

    def read_data(self) -> PlutusData:
        byts = self.read_bytes()

        return data_from_cbor(byts)

    def read_variable(self) -> Variable:
        index = self.read_integer(signed=False)

        return Variable(str(index))

    def read_lambda(self) -> Lambda:
        rhs = self.read_term()

        return Lambda("_", rhs)

    def read_apply(self) -> Apply:
        a = self.read_term()
        b = self.read_term()

        return Apply(a, b)

    def read_constant(self) -> Constant:
        type_list = self.read_linked_fixed_width_integer_list(
            self.tag_width("constType")
        )

        res = self.read_typed_value(type_list)

        return res

    def read_typed_value(self, type_list: List[int]) -> Constant:
        typed_reader = self.construct_typed_reader(type_list)

        assert len(type_list) == 0, "Did not consume all type parameters"

        return typed_reader()

    def sample_value(self, type_list: List[int]):
        typ = type_list.pop(0)
        if typ == 0:
            return BuiltinInteger(0)
        elif typ == 1:
            return BuiltinByteString(b"")
        elif typ == 2:
            return BuiltinString("")
        elif typ == 3:
            return BuiltinUnit()
        elif typ == 4:
            return BuiltinBool(False)
        elif typ in (5, 6):
            raise ValueError("unexpected type tag without type application")
        elif typ == 7:
            container_type = type_list.pop(0)
            if container_type == 5:
                list_type = self.sample_value(type_list)
                return BuiltinList([], list_type)
            else:
                assert container_type == 7, "Unexpected type tag"
                container_type = type_list.pop(0)
                if container_type == 6:
                    return BuiltinPair(
                        self.sample_value(type_list), self.sample_value(type_list)
                    )
        elif typ == 8:
            return PlutusInteger(0)
        else:
            raise ValueError(f"unhandled constant type {typ}")

    def construct_typed_reader(self, type_list: List[int]) -> Callable[[], Constant]:
        typ = type_list.pop(0)

        if typ == 0:
            return lambda: BuiltinInteger(self.read_integer(signed=True))
        elif typ == 1:
            return lambda: self.read_builtin_byte_string()
        elif typ == 2:
            return lambda: self.read_builtin_string()
        elif typ == 3:
            return lambda: BuiltinUnit()
        elif typ == 4:
            return lambda: BuiltinBool(self.read_bit() == 1)
        elif typ in (5, 6):
            raise ValueError("unexpected type tag without type application")
        elif typ == 7:
            container_type = type_list.pop(0)
            if container_type == 5:
                list_type = self.sample_value(type_list.copy())
                type_reader = self.construct_typed_reader(type_list)

                return lambda: BuiltinList(self.read_list(type_reader), list_type)
            else:
                assert container_type == 7, "Unexpected type tag"
                container_type = type_list.pop(0)
                if container_type == 6:
                    left_reader = self.construct_typed_reader(type_list)
                    right_reader = self.construct_typed_reader(type_list)
                    return lambda: BuiltinPair(left_reader(), right_reader())
                else:
                    raise ValueError(f"unhandled container type {container_type}")
        elif typ == 8:
            return lambda: self.read_data()
        else:
            raise ValueError(f"unhandled constant type {typ}")

    def read_delay(self) -> Delay:
        expr = self.read_term()

        return Delay(expr)

    def read_force(self) -> Force:
        expr = self.read_term()

        return Force(expr)

    def read_builtin(self) -> BuiltIn:
        id = self.read_tag("builtin")

        builtin = self.built_in_fun(id)

        return BuiltIn(builtin)

    def finalize(self):
        self.move_to_byte_boundary(True)

    def read_bits(self, num: int) -> str:
        bits = self._bits[self._pos : self._pos + num]
        self._pos += num
        return bits

    def read_fixed_width_integer(self, width: int) -> int:
        return int(self.read_bits(width), 2)

    def read_tag(self, name: str) -> int:
        return self.read_fixed_width_integer(self.tag_width(name))

    def read_bit(self) -> int:
        return self.read_fixed_width_integer(1)

    def read_byte(self) -> int:
        return self.read_fixed_width_integer(8)

    def read_program(self) -> Program:
        version = (
            self.read_integer(signed=False),
            self.read_integer(signed=False),
            self.read_integer(signed=False),
        )

        expr = self.read_term()

        self.finalize()

        return Program(version, expr)



import argparse
import enum
import json
import pathlib
import sys

import cbor2
import pycardano

from uplc.compiler_config import (
    DEFAULT_CONFIG,
    OPT_CONFIGS,
    ARGPARSE_ARGS,
    CompilationConfig,
)
from .tools import *
from .ast import Program, Apply
from .transformer import unique_variables
from .cost_model import (
    parse_builtin_cost_model,
    parse_cek_machine_cost_model,
    Budget,
    load_network_config,
    latest_network_config,
    updated_cek_machine_cost_model_from_network_config,
    updated_builtin_cost_model_from_network_config,
)


class Command(enum.Enum):
    eval = "eval"
    parse = "parse"
    build = "build"
    dump = "dump"


def get_args():
    a = argparse.ArgumentParser(
        description="An evaluator and compiler for UPLC written in python."
    )
    a.add_argument(
        "command",
        type=str,
        choices=Command.__members__.keys(),
        help="The command to execute on the input file.",
    )
    a.add_argument(
        "input_file",
        type=str,
        help="The input program to parse (foo.uplc). Set to - for stdin.",
    )
    a.add_argument(
        "-o",
        "--output-directory",
        default="",
        type=str,
        help="The output directory for artefacts of the build command. Defaults to the filename of the compiled contract. of the compiled contract.",
    )
    a.add_argument(
        "--dialect",
        default=UPLCDialect.Plutus.value,
        help="The dialect for dumping the parsed UPLC.",
        choices=[d.value for d in UPLCDialect],
    )
    a.add_argument(
        "--from-cbor",
        action="store_true",
        help="Read hex representation of flattened UPLC, wrapped in cbor.",
    )
    a.add_argument(
        "--from-hex",
        action="store_true",
        help="Read hex representation of flattened UPLC, not wrapped in cbor.",
    )
    a.add_argument(
        "args",
        nargs="*",
        default=[],
        help="Input parameters for the function, in case the command is eval.",
    )
    a.add_argument(
        "--recursion-limit",
        default=sys.getrecursionlimit(),
        help="Modify the recursion limit (necessary for larger UPLC programs)",
        type=int,
    )
    a.add_argument(
        "--builtin-cost-model-file",
        default=None,
        help="Provide a builtin cost-model file for the eval command. You will usually not need this.",
        type=str,
    )
    a.add_argument(
        "--cek-machine-cost-model-file",
        default=None,
        help="Provide a builtin cost-model file for the eval command. You will usually not need this.",
        type=str,
    )
    a.add_argument(
        "--cost-model-network-config",
        default=None,
        help="Provide a network config as propagated by the cardano-node for the eval command.",
        type=str,
    )
    a.add_argument(
        "--eval-cpu-budget",
        default=None,
        help="Provide a CPU budget for the eval command.",
        type=int,
    )
    a.add_argument(
        "--eval-memory-budget",
        default=None,
        help="Provide a Memory budget for the eval command.",
        type=int,
    )
    a.add_argument(
        "--plutus-version",
        default=2,
        help="Plutus version to use.",
        choices=[1, 2],
        type=int,
    )
    for k, v in ARGPARSE_ARGS.items():
        alts = v.pop("__alts__", [])
        a.add_argument(
            f"-f{k.replace('_', '-')}",
            *alts,
            **v,
            action="store_true",
            dest=k,
            default=None,
        )
        a.add_argument(
            f"-fno-{k.replace('_', '-')}",
            action="store_false",
            help=argparse.SUPPRESS,
            dest=k,
            default=None,
        )
    a.add_argument(
        "-O",
        default=1,
        type=int,
        help="The optimization level to use. Choose between 0 (nothing) and 3 (aggressive, semantics changing). Defaults to 1.",
        choices=range(len(OPT_CONFIGS)),
        dest="opt_level",
    )
    return a.parse_args()


def main():
    args = get_args()
    sys.setrecursionlimit(args.recursion_limit)

    # generate the compiler config
    compiler_config = DEFAULT_CONFIG
    compiler_config = compiler_config.update(OPT_CONFIGS[args.opt_level])
    overrides = {}
    for k in ARGPARSE_ARGS.keys():
        if getattr(args, k) is not None:
            overrides[k] = getattr(args, k)
    compiler_config = compiler_config.update(CompilationConfig(**overrides))

    command = Command(args.command)
    input_file = pathlib.Path(args.input_file) if args.input_file != "-" else sys.stdin
    with open(input_file, "r") as f:
        source_code = f.read()

    if args.from_cbor:
        code = unflatten(bytes.fromhex(source_code))
    elif args.from_hex:
        code = unflatten(cbor2.dumps(bytes.fromhex(source_code)))
    else:
        code: Program = parse(
            source_code,
            input_file.absolute() if isinstance(input_file, pathlib.Path) else None,
        )

    if command == Command.parse:
        print("Parsed successfully.")
        return

    code = compile(code, compiler_config)

    # Apply CLI parameters to code (i.e. to parameterize a parameterized contract)
    # UPLC lambdas may only take one argument at a time, so we evaluate by repeatedly applying
    code = apply(code, *map(lambda a: parse(f"(program 1.0.0 {a})").term, args.args))

    if command == Command.dump:
        print(dumps(code, UPLCDialect(args.dialect)))
        return

    if command == Command.build:
        if args.output_directory == "":
            if args.input_file == "-":
                print(
                    "Please supply an output directory if no input file is specified."
                )
                exit(-1)
            target_dir = pathlib.Path("build") / input_file.stem
        else:
            target_dir = pathlib.Path(args.output_directory)
        target_dir.mkdir(exist_ok=True, parents=True)
        cbor_bytes = flatten(code)
        cbor_hex = cbor_bytes.hex()
        # create cbor file for use with pycardano/lucid
        with (target_dir / "script.cbor").open("w") as fp:
            fp.write(cbor_hex)
        cbor = bytes.fromhex(cbor_hex)
        # double wrap
        cbor_wrapped = cbor2.dumps(cbor)
        cbor_wrapped_hex = cbor_wrapped.hex()
        # create plutus file
        d = {
            "type": "PlutusScriptV2",
            "description": f"",
            "cborHex": cbor_wrapped_hex,
        }
        with (target_dir / "script.plutus").open("w") as fp:
            json.dump(d, fp)
        script_hash = pycardano.plutus_script_hash(pycardano.PlutusV2Script(cbor))
        # generate policy ids
        with (target_dir / "script.policy_id").open("w") as fp:
            fp.write(script_hash.to_primitive().hex())
        addr_mainnet = pycardano.Address(
            script_hash, network=pycardano.Network.MAINNET
        ).encode()
        # generate addresses
        with (target_dir / "mainnet.addr").open("w") as fp:
            fp.write(addr_mainnet)
        addr_testnet = pycardano.Address(
            script_hash, network=pycardano.Network.TESTNET
        ).encode()
        with (target_dir / "testnet.addr").open("w") as fp:
            fp.write(addr_testnet)

        print(f"Wrote script artifacts to {target_dir}/")
        return
    if command == Command.eval:
        print("Starting execution")
        if args.builtin_cost_model_file is not None:
            with open(args.builtin_cost_model_file, "r") as fp:
                builtin_cost_model = parse_builtin_cost_model(json.load(fp))
        else:
            builtin_cost_model = default_builtin_cost_model_plutus_v2()
        if args.cek_machine_cost_model_file is not None:
            with open(args.cek_machine_cost_model_file, "r") as fp:
                cek_machine_cost_model = parse_cek_machine_cost_model(json.load(fp))
        else:
            cek_machine_cost_model = default_cek_machine_cost_model_plutus_v2()
        if args.cost_model_network_config is not None:
            with open(args.cost_model_network_config, "r") as fp:
                network_config = load_network_config(json.load(fp))
        else:
            network_config = latest_network_config()
        network_config = network_config[f"PlutusV{args.plutus_version}"]
        cek_machine_cost_model = updated_cek_machine_cost_model_from_network_config(
            cek_machine_cost_model, network_config
        )
        builtin_cost_model = updated_builtin_cost_model_from_network_config(
            builtin_cost_model, network_config
        )
        budget = default_budget()
        if args.eval_cpu_budget:
            budget.cpu = args.eval_cpu_budget
        if args.eval_memory_budget:
            budget.memory = args.eval_memory_budget
        ret = eval(
            code,
            budget=budget,
            cek_machine_cost_model=cek_machine_cost_model,
            builtin_cost_model=builtin_cost_model,
        )
        print("-------LOGS-------")
        if ret.logs:
            for line in ret.logs:
                print(line)
        else:
            print("None.")
        print("-------COST-------")
        print(f"CPU: {ret.cost.cpu}")
        print(f"Memory: {ret.cost.memory}")
        if isinstance(ret.result, Exception):
            print("-----ERROR-------")
            print(ret.result)
        else:
            print("-----SUCCESS-----")
            print(ret.result.dumps())


if __name__ == "__main__":
    main()



from copy import copy

from ..util import NodeTransformer
from ..ast import *


class FreeVariableError(ValueError):
    pass


class DeBrujinVariableTransformer(NodeTransformer):
    def __init__(self):
        self.scope = []

    def get_index(self, name: str):
        for index, var in enumerate(reversed(self.scope), start=1):
            if var == name:
                return index
        raise FreeVariableError(f"Variable {name} is never assigned")

    def visit_Lambda(self, node: Lambda):
        nc = copy(node)
        self.scope.append(node.var_name)
        nc.term = self.visit(node.term)
        self.scope.pop()
        return nc

    def visit_Variable(self, node: Variable):
        nc = copy(node)
        nc.name = self.get_index(node.name)
        return nc



import dataclasses
import enum
import json
import logging
import math
import typing
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum, auto
import hashlib
from typing import List, Any, Dict, Union

import cbor2
import frozendict
from frozenlist2 import frozenlist
import nacl.exceptions
from _cbor2 import CBOREncoder
from pycardano.crypto.bip32 import BIP32ED25519PublicKey
from pycardano.serialization import IndefiniteFrozenList, IndefiniteList

try:
    import pysecp256k1
except ImportError:
    pysecp256k1 = None

try:
    import pysecp256k1.extrakeys
    import pysecp256k1.schnorrsig as schnorrsig
except (RuntimeError, ImportError):
    schnorrsig = None


class UPLCDialect(enum.Enum):
    LegacyAiken = "legacy-aiken"
    Plutus = "plutus"


class Context:
    pass


@dataclass
class FrameApplyFun(Context):
    val: Any
    ctx: Context


@dataclass
class FrameApplyArg(Context):
    env: frozendict.frozendict
    term: "AST"
    ctx: Context


@dataclass
class FrameForce(Context):
    ctx: Context


@dataclass
class NoFrame(Context):
    pass


class Step:
    pass


@dataclass
class Return:
    context: Context
    value: Any


@dataclass
class Compute:
    ctx: Context
    env: frozendict.frozendict
    term: "AST"


@dataclass
class Done:
    term: "AST"


_LOGGER = logging.getLogger(__name__)


class AST:
    _fields = []

    def dumps(self, dialect=UPLCDialect.Plutus) -> str:
        raise NotImplementedError()

    def ex_mem(self) -> int:
        """The memory consumption of this element"""
        raise NotImplementedError()


@dataclass(frozen=True)
class Constant(AST):
    def dumps(self, dialect=UPLCDialect.Plutus) -> str:
        return f"(con {self.typestring(dialect=dialect)} {self.valuestring(dialect=dialect)})"

    def valuestring(self, dialect=UPLCDialect.Plutus):
        raise NotImplementedError()

    def typestring(self, dialect=UPLCDialect.Plutus):
        raise NotImplementedError()


@dataclass(frozen=True)
class BuiltinUnit(Constant):
    def typestring(self, dialect=UPLCDialect.Plutus):
        return "unit"

    def valuestring(self, dialect=UPLCDialect.Plutus):
        return "()"

    def ex_mem(self) -> int:
        return 1


@dataclass(frozen=True)
class BuiltinBool(Constant):
    value: bool

    def typestring(self, dialect=UPLCDialect.Plutus):
        return "bool"

    def valuestring(self, dialect=UPLCDialect.Plutus):
        return str(self.value)

    def ex_mem(self) -> int:
        return 1


@dataclass(frozen=True)
class BuiltinInteger(Constant):
    value: int

    def typestring(self, dialect=UPLCDialect.Plutus):
        return "integer"

    def valuestring(self, dialect=UPLCDialect.Plutus):
        return str(self.value)

    def ex_mem(self) -> int:
        if self.value == 0:
            return 1
        return (math.ceil(math.log2(abs(self.value))) // 64) + 1

    def __add__(self, other):
        assert isinstance(
            other, BuiltinInteger
        ), "Trying to add two non-builtin-integers"
        return BuiltinInteger(self.value + other.value)

    def __sub__(self, other):
        assert isinstance(
            other, BuiltinInteger
        ), "Trying to sub two non-builtin-integers"
        return BuiltinInteger(self.value - other.value)

    def __mul__(self, other):
        assert isinstance(
            other, BuiltinInteger
        ), "Trying to mul two non-builtin-integers"
        return BuiltinInteger(self.value * other.value)

    def __floordiv__(self, other):
        assert isinstance(
            other, BuiltinInteger
        ), "Trying to floordiv two non-builtin-integers"
        return BuiltinInteger(self.value // other.value)

    def __mod__(self, other):
        assert isinstance(
            other, BuiltinInteger
        ), "Trying to mod two non-builtin-integers"
        return BuiltinInteger(self.value % other.value)

    def __neg__(self):
        return BuiltinInteger(-self.value)

    def __eq__(self, other):
        assert isinstance(
            other, BuiltinInteger
        ), "Trying to eq two non-builtin-integers"
        return BuiltinBool(self.value == other.value)

    def __le__(self, other):
        assert isinstance(
            other, BuiltinInteger
        ), "Trying to le two non-builtin-integers"
        return BuiltinBool(self.value <= other.value)

    def __lt__(self, other):
        assert isinstance(
            other, BuiltinInteger
        ), "Trying to lt two non-builtin-integers"
        return BuiltinBool(self.value < other.value)


@dataclass(frozen=True)
class BuiltinByteString(Constant):
    value: bytes

    def typestring(self, dialect=UPLCDialect.Plutus):
        return "bytestring"

    def valuestring(self, dialect=UPLCDialect.Plutus):
        return f"#{self.value.hex()}"

    def ex_mem(self) -> int:
        if not self.value:
            return 1
        return ((len(self.value) - 1) // 8) + 1

    def __add__(self, other):
        assert isinstance(
            other, BuiltinByteString
        ), "Trying to add two non-builtin-bytestrings"
        return BuiltinByteString(self.value + other.value)

    def __len__(self):
        return BuiltinInteger(len(self.value))

    def __eq__(self, other):
        assert isinstance(
            other, BuiltinByteString
        ), "Trying to eq two non-builtin-bytestrings"
        return BuiltinBool(self.value == other.value)

    def __le__(self, other):
        assert isinstance(
            other, BuiltinByteString
        ), "Trying to le two non-builtin-bytestrings"
        return BuiltinBool(self.value <= other.value)

    def __lt__(self, other):
        assert isinstance(
            other, BuiltinByteString
        ), "Trying to lt two non-builtin-bytestrings"
        return BuiltinBool(self.value < other.value)

    def __getitem__(self, item):
        if isinstance(item, BuiltinInteger):
            assert 0 <= item.value <= len(self.value), "Out of bounds"
            return BuiltinInteger(self.value[item.value])
        raise NotImplementedError()

    def decode(self, *args):
        return BuiltinString(self.value.decode("utf8"))


@dataclass(frozen=True)
class BuiltinString(Constant):
    value: str

    def typestring(self, dialect=UPLCDialect.Plutus):
        return "string"

    def valuestring(self, dialect=UPLCDialect.Plutus):
        return json.dumps(self.value)

    def ex_mem(self) -> int:
        return len(self.value)

    def __add__(self, other):
        assert isinstance(other, BuiltinString), "Can only add two bytestrings"
        return BuiltinString(self.value + other.value)

    def __eq__(self, other):
        assert isinstance(
            other, BuiltinString
        ), "Can only compare two bytestrings for equality"
        return BuiltinBool(self.value == other.value)

    def encode(self, *args):
        return BuiltinByteString(self.value.encode())


@dataclass(frozen=True)
class BuiltinPair(Constant):
    l_value: Constant
    r_value: Constant

    def typestring(self, dialect=UPLCDialect.Plutus):
        if dialect == UPLCDialect.LegacyAiken:
            return f"pair<{self.l_value.typestring(dialect=dialect)}, {self.r_value.typestring(dialect=dialect)}>"
        return f"(pair {self.l_value.typestring(dialect=dialect)} {self.r_value.typestring(dialect=dialect)})"

    def valuestring(self, dialect=UPLCDialect.Plutus):
        if dialect == UPLCDialect.LegacyAiken:
            return f"[{self.l_value.valuestring(dialect=dialect)}, {self.r_value.valuestring(dialect=dialect)}]"
        return f"({self.l_value.valuestring(dialect=dialect)}, {self.r_value.valuestring(dialect=dialect)})"

    def ex_mem(self) -> int:
        return self.l_value.ex_mem() + self.r_value.ex_mem()

    def __getitem__(self, item):
        if isinstance(item, int):
            if item == 0:
                return self.l_value
            elif item == 1:
                return self.r_value
        raise ValueError()


@dataclass(frozen=True, init=False)
class BuiltinList(Constant):
    values: List[Constant]
    # dirty hack to handle the type of empty lists
    sample_value: Constant

    def __init__(self, values, sample_value=None):
        object.__setattr__(self, "values", frozenlist(values))
        if not values:
            assert (
                sample_value is not None
            ), "Need to provide a sample value for empty lists to infer the type"
            object.__setattr__(self, "sample_value", sample_value)
        else:
            object.__setattr__(self, "sample_value", values[0])

    def typestring(self, dialect=UPLCDialect.Plutus):
        if dialect == UPLCDialect.LegacyAiken:
            return f"list<{self.sample_value.typestring(dialect=dialect)}>"
        return f"(list {self.sample_value.typestring(dialect=dialect)})"

    def valuestring(self, dialect=UPLCDialect.Plutus):
        return f"[{', '.join(v.valuestring(dialect=dialect) for v in self.values)}]"

    def ex_mem(self) -> int:
        return sum(v.ex_mem() for v in self.values)

    def __add__(self, other):
        assert isinstance(other, BuiltinList), "Can only append two lists"
        assert (
            other.typestring() == self.typestring()
        ), f"Expected {self.typestring()} but got {other.typestring()}"
        return BuiltinList(self.values + other.values)

    def __eq__(self, other):
        assert isinstance(other, BuiltinList), "Can only compare two lists"
        return self.values == other.values

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.values[item]
        elif isinstance(item, slice):
            return BuiltinList(self.values[item], self.sample_value)


@dataclass(frozen=True)
class PlutusData(Constant):
    def dumps(self, dialect=UPLCDialect.Plutus):
        internal_string = self.valuestring(dialect=dialect)
        if dialect == UPLCDialect.Plutus:
            internal_string = f"({internal_string})"
        return f"(con data {internal_string})"

    def valuestring(self, dialect=UPLCDialect.Plutus) -> str:
        if dialect == UPLCDialect.LegacyAiken:
            return f"#{plutus_cbor_dumps(self).hex()}"
        return f"{self.plutus_valuestring()}"

    def plutus_valuestring(self):
        raise NotImplementedError

    def typestring(self, dialect=UPLCDialect.Plutus):
        return "data"

    def to_cbor(self) -> Any:
        """Returns a CBOR encodable representation of this object"""
        raise NotImplementedError

    def to_json(self) -> dict:
        """Returns a JSON encodable representation of this object"""
        raise NotImplementedError

    def ex_mem(self) -> int:
        return 4 + self.d_ex_mem()

    def d_ex_mem(self) -> int:
        """Ex-Mem without the constant 4 cost for deconstruction"""
        raise NotImplementedError()


@dataclass(frozen=True)
class PlutusAtomic(PlutusData):
    value: Any

    def to_cbor(self):
        return self


@dataclass(frozen=True, eq=True)
class PlutusInteger(PlutusAtomic):
    value: int

    def to_json(self):
        return {"int": self.value}

    def plutus_valuestring(self):
        return f"I {self.value}"

    def d_ex_mem(self) -> int:
        return BuiltinInteger(self.value).ex_mem()


@dataclass(frozen=True, eq=True)
class PlutusByteString(PlutusAtomic):
    value: bytes

    def to_json(self):
        return {"bytes": self.value.hex()}

    def plutus_valuestring(self):
        return f"B #{self.value.hex()}"

    def d_ex_mem(self) -> int:
        return BuiltinByteString(self.value).ex_mem()


@dataclass(frozen=True, eq=True)
class PlutusList(PlutusData):
    value: Union[List[PlutusData], frozenlist]

    def __post_init__(self):
        object.__setattr__(self, "value", frozenlist(self.value))

    def to_cbor(self):
        return [d.to_cbor() for d in self.value]

    def to_json(self):
        return {"list": [v.to_json() for v in self.value]}

    def plutus_valuestring(self):
        return f"List [{', '.join(x.plutus_valuestring() for x in self.value)}]"

    def d_ex_mem(self) -> int:
        return sum(v.ex_mem() for v in self.value)


@dataclass(frozen=True, eq=True)
class PlutusMap(PlutusData):
    value: Union[Dict[PlutusData, PlutusData], frozendict.frozendict]

    def __post_init__(self):
        frozen_value = frozendict.frozendict(self.value)
        object.__setattr__(self, "value", frozen_value)

    def to_cbor(self):
        return {k.to_cbor(): v.to_cbor() for k, v in self.value.items()}

    def to_json(self):
        return {
            "map": [{"k": k.to_json(), "v": v.to_json()} for k, v in self.value.items()]
        }

    def plutus_valuestring(self):
        recursive_val_strings = (
            f"({x.plutus_valuestring()}, {y.plutus_valuestring()})"
            for x, y in self.value.items()
        )
        return f"Map [{', '.join(recursive_val_strings)}]"

    def d_ex_mem(self) -> int:
        return sum(v.ex_mem() + k.ex_mem() for k, v in self.value.items())


@dataclass(frozen=True, eq=True)
class PlutusConstr(PlutusData):
    constructor: int
    fields: Union[List[PlutusData], frozenlist]

    def __post_init__(self):
        object.__setattr__(self, "fields", frozenlist(self.fields))

    def to_cbor(self):
        if self.fields:
            fields = IndefiniteFrozenList([f.to_cbor() for f in self.fields])
            fields.freeze()
        else:
            fields = []
        if 0 <= self.constructor < 7:
            return cbor2.CBORTag(self.constructor + 121, fields)
        elif 7 <= self.constructor < 128:
            return cbor2.CBORTag((self.constructor - 7) + 1280, fields)
        else:
            return cbor2.CBORTag(102, [self.constructor, fields])

    def to_json(self):
        return {
            "constructor": self.constructor,
            "fields": [v.to_json() for v in self.fields],
        }

    def plutus_valuestring(self):
        return f"Constr {self.constructor} [{', '.join(x.plutus_valuestring() for x in self.fields)}]"

    def d_ex_mem(self) -> int:
        return sum(v.ex_mem() for v in self.fields)


def _int_to_bytes(x: int):
    return x.to_bytes((x.bit_length() + 7) // 8, byteorder="big")


def default_encoder(encoder: CBOREncoder, value: Union[PlutusData, IndefiniteList]):
    """A fallback function that encodes PlutusData objects"""
    if isinstance(value, IndefiniteList):
        # Currently, cbor2 doesn't support indefinite list, therefore we need special
        # handling here to explicitly write header (b'\x9f'), each body item, and footer (b'\xff') to
        # the output bytestring.
        encoder.write(b"\x9f")
        for item in value:
            encoder.encode(item)
        encoder.write(b"\xff")
        return
    if not isinstance(value, PlutusData):
        raise NotImplementedError(f"Can not encode type {type(value)}")
    value = value.to_cbor()
    if isinstance(value, PlutusByteString):
        # the encoder can not handle indefinite length arrays, but the plutus standard
        # requires encoding bytes as indefinite byte sequence where each chunk is at most 64 bytes long
        byts = value.value
    elif isinstance(value, PlutusInteger):
        if -(2**64) < value.value < 2**64 - 1:
            encoder.encode(value.value)
            return
        if value.value >= 0:
            byts = _int_to_bytes(value.value)
            encoder.write(b"\xc2")
        else:
            byts = _int_to_bytes(-value.value - 1)
            encoder.write(b"\xc3")
    else:
        encoder.encode(value)
        return
    if len(byts) < 64:
        encoder.encode(byts)
        return
    encoder.write(b"\x5f")
    max_chunk_len = 64
    n = len(byts)
    pos = 0
    while pos < n:
        n_chunk = min(n - pos, max_chunk_len)
        chunk = byts[pos : pos + n_chunk]
        encoder.encode(chunk)
        pos += n_chunk
    encoder.write(b"\xff")


def plutus_cbor_dumps(x):
    return cbor2.dumps(x, default=default_encoder)


def data_from_cbortag(cbor) -> PlutusData:
    if isinstance(cbor, cbor2.CBORTag):
        if 121 <= cbor.tag <= 121 + 6:
            constructor = cbor.tag - 121
            fields = cbor.value
        elif 1280 <= cbor.tag <= 1280 + (127 - 7):
            constructor = cbor.tag - 1280 + 7
            fields = cbor.value
        elif cbor.tag == 102:
            constructor, fields = cbor.value
        else:
            raise ValueError(f"Invalid cbor with tag {cbor.tag}")
        fields = frozenlist(list(map(data_from_cbortag, fields)))
        return PlutusConstr(constructor, fields)
    if isinstance(cbor, int):
        return PlutusInteger(cbor)
    if isinstance(cbor, bytes):
        return PlutusByteString(cbor)
    if isinstance(cbor, list):
        entries = frozenlist(list(map(data_from_cbortag, cbor)))
        return PlutusList(entries)
    if isinstance(cbor, dict):
        return PlutusMap(
            frozendict.frozendict(
                {data_from_cbortag(k): data_from_cbortag(v) for k, v in cbor.items()}
            )
        )
    raise NotImplementedError(f"Unknown cbor type notation in {cbor}")


def data_from_cbor(cbor: bytes) -> PlutusData:
    raw_datum = cbor2.loads(cbor)
    return data_from_cbortag(raw_datum)


def data_from_json_dict(d: dict) -> PlutusData:
    if "constructor" in d:
        fields = frozenlist([data_from_json_dict(f) for f in d["fields"]])
        return PlutusConstr(d["constructor"], fields)
    if "int" in d:
        return PlutusInteger(d["int"])
    if "bytes" in d:
        return PlutusByteString(bytes.fromhex(d["bytes"]))
    if "list" in d:
        entries = frozenlist(list(map(data_from_json_dict, d["list"])))
        return PlutusList(entries)
    if "map" in d:
        return PlutusMap(
            frozendict.frozendict(
                {
                    data_from_json_dict(m["k"]): data_from_json_dict(m["v"])
                    for m in d["map"]
                }
            )
        )
    raise NotImplementedError(f"Unknown json notation in {d}")


def data_from_json(json_string: str) -> PlutusData:
    raw_datum = json.loads(json_string)
    return data_from_json_dict(raw_datum)


def plutus_json_dumps(x: PlutusData):
    return json.dumps(x.to_json())


class ConstantType(Enum):
    integer = auto()
    bytestring = auto()
    string = auto()
    unit = auto()
    bool = auto()
    pair = auto()
    list = auto()
    data = auto()


# As found in https://plutonomicon.github.io/plutonomicon/builtin-functions
# NOTE it is crucial that the values matches table C.3 in the plutus core spec
# https://ci.iog.io/build/1230997/download/1/plutus-core-specification.pdf
class BuiltInFun(Enum):
    # Integers
    AddInteger = 0
    SubtractInteger = 1
    MultiplyInteger = 2
    DivideInteger = 3
    QuotientInteger = 4
    RemainderInteger = 5
    ModInteger = 6
    EqualsInteger = 7
    LessThanInteger = 8
    LessThanEqualsInteger = 9
    # Bytestrings
    AppendByteString = 10
    ConsByteString = 11
    SliceByteString = 12
    LengthOfByteString = 13
    IndexByteString = 14
    EqualsByteString = 15
    LessThanByteString = 16
    LessThanEqualsByteString = 17
    # Cryptography and hashes
    Sha2_256 = 18
    Sha3_256 = 19
    Blake2b_256 = 20
    # Keccak_256 = 71
    # Blake2b_224 = 72
    VerifyEd25519Signature = 21  # formerly verifySignature
    VerifyEcdsaSecp256k1Signature = 52
    VerifySchnorrSecp256k1Signature = 53
    # Strings
    AppendString = 22
    EqualsString = 23
    EncodeUtf8 = 24
    DecodeUtf8 = 25
    # Bool
    IfThenElse = 26
    # Unit
    ChooseUnit = 27
    # Tracing
    Trace = 28
    # Pairs
    FstPair = 29
    SndPair = 30
    # Lists
    ChooseList = 31
    MkCons = 32
    HeadList = 33
    TailList = 34
    NullList = 35
    # Data
    ChooseData = 36
    ConstrData = 37
    MapData = 38
    ListData = 39
    IData = 40
    BData = 41
    UnConstrData = 42
    UnMapData = 43
    UnListData = 44
    UnIData = 45
    UnBData = 46
    EqualsData = 47
    SerialiseData = 51
    # Misc monomorphized constructors
    MkPairData = 48
    MkNilData = 49
    MkNilPairData = 50
    # BLS Builtins
    # Bls12_381_G1_Add = 54
    # Bls12_381_G1_Neg = 55
    # Bls12_381_G1_ScalarMul = 56
    # Bls12_381_G1_Equal = 57
    # Bls12_381_G1_Compress = 58
    # Bls12_381_G1_Uncompress = 59
    # Bls12_381_G1_HashToGroup = 60
    # Bls12_381_G2_Add = 61
    # Bls12_381_G2_Neg = 62
    # Bls12_381_G2_ScalarMul = 63
    # Bls12_381_G2_Equal = 64
    # Bls12_381_G2_Compress = 65
    # Bls12_381_G2_Uncompress = 66
    # Bls12_381_G2_HashToGroup = 67
    # Bls12_381_MillerLoop = 68
    # Bls12_381_MulMlResult = 69
    # Bls12_381_FinalVerify = 70


def typechecked(*typs):
    def typecheck_decorator(fun):
        if len(typs) == 1:

            def wrapped_fun(a1):
                for i, (arg, typ) in enumerate(zip([a1], typs)):
                    assert isinstance(
                        arg, typ
                    ), f"Argument {i} has invalid type, expected type {typ} got {type(arg)} ({arg})"
                return fun(a1)

        elif len(typs) == 2:

            def wrapped_fun(a1, a2):
                for i, (arg, typ) in enumerate(zip([a1, a2], typs)):
                    assert isinstance(
                        arg, typ
                    ), f"Argument {i} has invalid type, expected type {typ} got {type(arg)} ({arg})"
                return fun(a1, a2)

        elif len(typs) == 3:

            def wrapped_fun(a1, a2, a3):
                for i, (arg, typ) in enumerate(zip([a1, a2, a3], typs)):
                    assert isinstance(
                        arg, typ
                    ), f"Argument {i} has invalid type, expected type {typ} got {type(arg)} ({arg})"
                return fun(a1, a2, a3)

        elif len(typs) == 4:

            def wrapped_fun(a1, a2, a3, a4):
                for i, (arg, typ) in enumerate(zip([a1, a2, a3, a4], typs)):
                    assert isinstance(
                        arg, typ
                    ), f"Argument {i} has invalid type, expected type {typ} got {type(arg)} ({arg})"
                return fun(a1, a2, a3, a4)

        elif len(typs) == 5:

            def wrapped_fun(a1, a2, a3, a4, a5):
                for i, (arg, typ) in enumerate(zip([a1, a2, a3, a4, a5], typs)):
                    assert isinstance(
                        arg, typ
                    ), f"Argument {i} has invalid type, expected type {typ} got {type(arg)} ({arg})"
                return fun(a1, a2, a3, a4, a5)

        elif len(typs) == 6:

            def wrapped_fun(a1, a2, a3, a4, a5, a6):
                for i, (arg, typ) in enumerate(zip([a1, a2, a3, a4, a5, a6], typs)):
                    assert isinstance(
                        arg, typ
                    ), f"Argument {i} has invalid type, expected type {typ} got {type(arg)} ({arg})"
                return fun(a1, a2, a3, a4, a5, a6)

        else:
            raise NotImplementedError("Too many arguments")
        return wrapped_fun

    return typecheck_decorator


@typechecked(BuiltinBool, AST, AST)
def _IfThenElse(i, t, e):
    return t if i.value else e


@typechecked(PlutusData, AST, AST, AST, AST, AST)
def _ChooseData(d, v, w, x, y, z):
    if isinstance(d, PlutusConstr):
        return v
    if isinstance(d, PlutusMap):
        return w
    if isinstance(d, PlutusList):
        return x
    if isinstance(d, PlutusInteger):
        return y
    if isinstance(d, PlutusByteString):
        return z


@typechecked(BuiltinByteString, BuiltinByteString, BuiltinByteString)
def verify_ed25519(pk: BuiltinByteString, m: BuiltinByteString, s: BuiltinByteString):
    assert len(pk.value) == 32, "Ed25519S PublicKey should be 32 bytes"
    assert len(s.value) == 64, "Ed25519S Signature should be 64 bytes"
    try:
        BIP32ED25519PublicKey(pk.value[:32], pk.value[32:]).verify(s.value, m.value)
        return BuiltinBool(True)
    except nacl.exceptions.BadSignatureError:
        return BuiltinBool(False)


@typechecked(BuiltinByteString, BuiltinByteString, BuiltinByteString)
def verify_ecdsa_secp256k1(
    pk: BuiltinByteString, m: BuiltinByteString, s: BuiltinByteString
):
    # TODO length checks
    if pysecp256k1 is None:
        _LOGGER.error("libsecp256k1 is not installed. ECDSA verification will not work")
        raise RuntimeError("ECDSA not supported")
    pubkey = pysecp256k1.ec_pubkey_parse(pk.value)
    sig = pysecp256k1.ecdsa_signature_parse_compact(s.value)
    res = pysecp256k1.ecdsa_verify(sig, pubkey, m.value)
    return BuiltinBool(res)


@typechecked(BuiltinByteString, BuiltinByteString, BuiltinByteString)
def verify_schnorr_secp256k1(
    pk: BuiltinByteString, m: BuiltinByteString, s: BuiltinByteString
):
    # TODO length checks
    if pysecp256k1 is None:
        _LOGGER.error("libsecp256k1 is not installed. ECDSA verification will not work")
        raise RuntimeError("ECDSA not supported")
    if schnorrsig is None:
        _LOGGER.error(
            "libsecp256k1 is installed without schnorr support. Schnorr verification will not work"
        )
        raise RuntimeError("Schnorr not supported")
    pubkey = pysecp256k1.extrakeys.xonly_pubkey_parse(pk.value)
    res = schnorrsig.schnorrsig_verify(s.value, m.value, pubkey)
    return BuiltinBool(res)


def _quot(a, b):
    return a // b if (a * b > BuiltinInteger(0)).value else (a + (-a % b)) // b


@typechecked(BuiltinList)
def _TailList(xs: BuiltinList):
    if xs.values == []:
        raise RuntimeError("Can not tailList on an empty list")
    return xs[1:]


def _MkCons(x, xs):
    assert isinstance(xs, BuiltinList), "Can only cons onto a list"
    assert isinstance(x, xs.sample_value.__class__) or (
        isinstance(x, PlutusData) and (isinstance(xs.sample_value, PlutusData))
    ), "Can only cons elements of the same type"
    return BuiltinList([x]) + xs


def _MapData(x):
    assert isinstance(x, BuiltinList), "Can only map over a list"
    assert isinstance(x.sample_value, BuiltinPair), "Can only map over a list of pairs"
    return PlutusMap({p.l_value: p.r_value for p in x.values})


two_ints = typechecked(BuiltinInteger, BuiltinInteger)
two_bytestrings = typechecked(BuiltinByteString, BuiltinByteString)
two_strings = typechecked(BuiltinString, BuiltinString)
single_bytestring = typechecked(BuiltinByteString)
single_data = typechecked(PlutusData)
single_data_int = typechecked(PlutusInteger)
single_data_bytes = typechecked(PlutusByteString)
single_data_list = typechecked(PlutusList)
single_data_map = typechecked(PlutusMap)
single_data_constr = typechecked(PlutusConstr)

BuiltInFunEvalMap = {
    BuiltInFun.AddInteger: two_ints(lambda x, y: x + y),
    BuiltInFun.SubtractInteger: two_ints(lambda x, y: x - y),
    BuiltInFun.MultiplyInteger: two_ints(lambda x, y: x * y),
    # round towards -inf
    BuiltInFun.DivideInteger: two_ints(lambda x, y: x // y),
    # round towards 0
    BuiltInFun.QuotientInteger: two_ints(_quot),
    # (x `quot` y)*y + (x `rem` y) == x
    BuiltInFun.RemainderInteger: two_ints(lambda x, y: x - _quot(x, y) * y),
    # (x `div` y)*y + (x `mod` y) == x
    BuiltInFun.ModInteger: two_ints(lambda x, y: x % y),
    BuiltInFun.EqualsInteger: two_ints(lambda x, y: x == y),
    BuiltInFun.LessThanInteger: two_ints(lambda x, y: x < y),
    BuiltInFun.LessThanEqualsInteger: two_ints(lambda x, y: x <= y),
    BuiltInFun.AppendByteString: two_bytestrings(lambda x, y: x + y),
    BuiltInFun.ConsByteString: typechecked(BuiltinInteger, BuiltinByteString)(
        lambda x, y: BuiltinByteString(bytes([x.value])) + y
    ),
    BuiltInFun.SliceByteString: typechecked(
        BuiltinInteger, BuiltinInteger, BuiltinByteString
    )(lambda x, y, z: BuiltinByteString(z.value[max(x.value, 0) :][: max(y.value, 0)])),
    BuiltInFun.LengthOfByteString: single_bytestring(
        lambda x: BuiltinInteger(len(x.value))
    ),
    BuiltInFun.IndexByteString: typechecked(BuiltinByteString, BuiltinInteger)(
        lambda x, y: x[y]
    ),
    BuiltInFun.EqualsByteString: two_bytestrings(lambda x, y: x == y),
    BuiltInFun.LessThanByteString: two_bytestrings(lambda x, y: x < y),
    BuiltInFun.LessThanEqualsByteString: two_bytestrings(lambda x, y: x <= y),
    BuiltInFun.Sha2_256: single_bytestring(
        lambda x: BuiltinByteString(hashlib.sha256(x.value).digest())
    ),
    BuiltInFun.Sha3_256: single_bytestring(
        lambda x: BuiltinByteString(hashlib.sha3_256(x.value).digest())
    ),
    BuiltInFun.Blake2b_256: single_bytestring(
        lambda x: BuiltinByteString(hashlib.blake2b(x.value, digest_size=32).digest())
    ),
    # BuiltInFun.VerifySignature: verify_ed25519,
    BuiltInFun.VerifyEd25519Signature: verify_ed25519,
    BuiltInFun.VerifyEcdsaSecp256k1Signature: verify_ecdsa_secp256k1,
    BuiltInFun.VerifySchnorrSecp256k1Signature: verify_schnorr_secp256k1,
    BuiltInFun.AppendString: two_strings(lambda x, y: x + y),
    BuiltInFun.EqualsString: two_strings(lambda x, y: x == y),
    BuiltInFun.EncodeUtf8: typechecked(BuiltinString)(
        lambda x: BuiltinByteString(x.value.encode("utf8"))
    ),
    BuiltInFun.DecodeUtf8: single_bytestring(
        lambda x: BuiltinString(x.value.decode("utf8"))
    ),
    BuiltInFun.IfThenElse: _IfThenElse,
    BuiltInFun.ChooseUnit: typechecked(BuiltinUnit, AST)(lambda x, y: y),
    BuiltInFun.Trace: typechecked(BuiltinString, AST)(lambda x, y: y),
    BuiltInFun.FstPair: typechecked(BuiltinPair)(lambda x: x[0]),
    BuiltInFun.SndPair: typechecked(BuiltinPair)(lambda x: x[1]),
    BuiltInFun.ChooseList: typechecked(BuiltinList, AST, AST)(
        lambda l, x, y: x if BuiltinList([], l.sample_value) == l else y
    ),
    BuiltInFun.MkCons: _MkCons,
    BuiltInFun.HeadList: typechecked(BuiltinList)(lambda l: l[0]),
    BuiltInFun.TailList: _TailList,
    BuiltInFun.NullList: typechecked(BuiltinList)(
        lambda l: BuiltinBool(l.values == [])
    ),
    BuiltInFun.ChooseData: _ChooseData,
    BuiltInFun.ConstrData: typechecked(BuiltinInteger, BuiltinList)(
        lambda x, y: PlutusConstr(x.value, y.values)
    ),
    BuiltInFun.MapData: _MapData,
    BuiltInFun.ListData: typechecked(BuiltinList)(lambda x: PlutusList(x.values)),
    BuiltInFun.IData: typechecked(BuiltinInteger)(lambda x: PlutusInteger(x.value)),
    BuiltInFun.BData: single_bytestring(lambda x: PlutusByteString(x.value)),
    BuiltInFun.UnConstrData: single_data_constr(
        lambda x: BuiltinPair(
            BuiltinInteger(x.constructor), BuiltinList(x.fields, PlutusData())
        )
    ),
    BuiltInFun.UnMapData: single_data_map(
        lambda x: BuiltinList(
            [BuiltinPair(k, v) for k, v in x.value.items()],
            BuiltinPair(PlutusData(), PlutusData()),
        )
    ),
    BuiltInFun.UnListData: single_data_list(
        lambda x: BuiltinList(x.value, PlutusData())
    ),
    BuiltInFun.UnIData: single_data_int(lambda x: BuiltinInteger(x.value)),
    BuiltInFun.UnBData: single_data_bytes(lambda x: BuiltinByteString(x.value)),
    BuiltInFun.EqualsData: typechecked(PlutusData, PlutusData)(
        lambda x, y: BuiltinBool(x == y)
    ),
    BuiltInFun.MkPairData: typechecked(Constant, Constant)(
        lambda x, y: BuiltinPair(x, y)
    ),
    BuiltInFun.MkNilData: lambda _: BuiltinList([], PlutusData()),
    BuiltInFun.MkNilPairData: lambda _: BuiltinList(
        [], BuiltinPair(PlutusData(), PlutusData())
    ),
    BuiltInFun.SerialiseData: single_data(
        lambda x: BuiltinByteString(plutus_cbor_dumps(x))
    ),
}

BuiltInFunForceMap = defaultdict(int)
BuiltInFunForceMap.update(
    {
        BuiltInFun.IfThenElse: 1,
        BuiltInFun.ChooseUnit: 1,
        BuiltInFun.Trace: 1,
        BuiltInFun.FstPair: 2,
        BuiltInFun.SndPair: 2,
        BuiltInFun.ChooseList: 2,
        BuiltInFun.MkCons: 1,
        BuiltInFun.HeadList: 1,
        BuiltInFun.TailList: 1,
        BuiltInFun.NullList: 1,
        BuiltInFun.ChooseData: 1,
    }
)


@dataclass
class Program(AST):
    version: typing.Tuple[int, int, int]
    term: AST
    _fields = ["term"]

    def eval(self, context, state):
        return self.term.eval(context, state)

    def dumps(self, dialect=UPLCDialect.Plutus) -> str:
        return f"(program {'.'.join(str(x) for x in self.version)} {self.term.dumps(dialect=dialect)})"


@dataclass
class Variable(AST):
    name: str

    def dumps(self, dialect=UPLCDialect.Plutus) -> str:
        return self.name


@dataclass
class BoundStateLambda(AST):
    var_name: str
    term: AST
    state: frozendict.frozendict
    _fields = ["term"]

    def dumps(self, dialect=UPLCDialect.Plutus) -> str:
        s = f"(lam {self.var_name} {self.term.dumps(dialect=dialect)})"
        for k, v in reversed(self.state.items()):
            s = f"[(lam {k} {s}) {v.dumps(dialect=dialect)}]"
        return s

    def ex_mem(self) -> int:
        return 1


@dataclass
class Lambda(BoundStateLambda):
    var_name: str
    term: AST
    state: frozendict.frozendict = dataclasses.field(
        default_factory=frozendict.frozendict
    )


@dataclass
class BoundStateDelay(AST):
    term: AST
    state: frozendict.frozendict
    _fields = ["term"]

    def dumps(self, dialect=UPLCDialect.Plutus) -> str:
        s = f"(delay {self.term.dumps(dialect=dialect)})"
        for k, v in reversed(self.state.items()):
            s = f"[(lam {k} {s}) {v.dumps(dialect=dialect)}]"
        return s

    def ex_mem(self) -> int:
        return 1


@dataclass
class Delay(BoundStateDelay):
    term: AST
    state: frozendict.frozendict = dataclasses.field(
        default_factory=frozendict.frozendict
    )


@dataclass
class Force(AST):
    term: AST
    _fields = ["term"]

    def dumps(self, dialect=UPLCDialect.Plutus) -> str:
        return f"(force {self.term.dumps(dialect=dialect)})"


@dataclass
class ForcedBuiltIn(AST):
    builtin: BuiltInFun
    applied_forces: int
    bound_arguments: List[AST]

    def dumps(self, dialect=UPLCDialect.Plutus) -> str:
        if len(self.bound_arguments):
            return Apply(
                ForcedBuiltIn(
                    self.builtin, self.applied_forces, self.bound_arguments[:-1]
                ),
                self.bound_arguments[-1],
            ).dumps(dialect=dialect)
        if self.applied_forces > 0:
            return Force(
                ForcedBuiltIn(
                    self.builtin, self.applied_forces - 1, self.bound_arguments
                )
            ).dumps(dialect=dialect)
        return f"(builtin {self.builtin.name[0].lower()}{self.builtin.name[1:]})"

    def ex_mem(self) -> int:
        return 1


@dataclass
class BuiltIn(ForcedBuiltIn):
    builtin: BuiltInFun
    applied_forces: int = dataclasses.field(default=0)
    bound_arguments: list = dataclasses.field(default_factory=lambda: [])


@dataclass
class Error(AST):
    def dumps(self, dialect=UPLCDialect.Plutus) -> str:
        return f"(error)"


@dataclass
class Apply(AST):
    f: AST
    x: AST
    _fields = ["f", "x"]

    def dumps(self, dialect=UPLCDialect.Plutus) -> str:
        return f"[{self.f.dumps(dialect=dialect)} {self.x.dumps(dialect=dialect)}]"



from dataclasses import dataclass
from typing import Optional, Union


@dataclass(frozen=True)
class CompilationConfig:
    constant_folding: Optional[bool] = None
    unique_variable_names: Optional[bool] = None
    remove_traces: Optional[bool] = None
    remove_force_delay: Optional[bool] = None

    def update(
        self, other: Optional["CompilationConfig"] = None, **kwargs
    ) -> "CompilationConfig":
        own_dict = self.__dict__
        other_dict = other.__dict__ if isinstance(other, CompilationConfig) else kwargs
        return self.__class__(
            **{
                k: other_dict.get(k) if other_dict.get(k) is not None else own_dict[k]
                for k in own_dict
            }
        )


# The default configuration for the compiler
OPT_O0_CONFIG = CompilationConfig()
OPT_O1_CONFIG = OPT_O0_CONFIG.update(remove_force_delay=True)
OPT_O2_CONFIG = OPT_O1_CONFIG.update(constant_folding=True)
OPT_O3_CONFIG = OPT_O2_CONFIG.update(remove_traces=True)
OPT_CONFIGS = [OPT_O0_CONFIG, OPT_O1_CONFIG, OPT_O2_CONFIG, OPT_O3_CONFIG]

DEFAULT_CONFIG = CompilationConfig(unique_variable_names=False).update(OPT_O1_CONFIG)

ARGPARSE_ARGS = {
    "remove_traces": {
        "help": "Remove traces from the compiled contract.",
    },
    "unique_variable_names": {
        "__alts__": ["--unique-varnames"],
        "help": "Assign variables a unique name.",
    },
    "constant_folding": {
        "__alts__": ["--cf"],
        "help": "Enables experimental constant folding, including propagation and code execution.",
    },
    "remove_force_delay": {
        "__alts__": ["--rfd"],
        "help": "Removes delayed terms that are immediately forced.",
    },
}
for k in ARGPARSE_ARGS:
    assert (
        k in DEFAULT_CONFIG.__dict__
    ), f"Key {k} not found in CompilationConfig.__dict__"



from ..util import NodeTransformer, NodeVisitor
from ..ast import *

"""
Optimizes code by pre-evaluating each subterm
If it throws an error, assume it is not safe to pre-evaluate and don't replace, otherwise replace by result.
"""


class TraceFinder(NodeVisitor):
    found = False

    def visit_BuiltIn(self, node: BuiltIn):
        return self.visit_ForcedBuiltIn(node)

    def visit_ForcedBuiltIn(self, node: ForcedBuiltIn):
        if node.builtin == BuiltInFun.Trace:
            self.found = True
            return
        return

    def contains_trace(self, node: AST) -> bool:
        self.visit(node)
        return self.found


_CONSTANT_FINAL_RESULTS = (
    Constant,
    Variable,
    Error,
)


class PreEvaluationOptimizer(NodeTransformer):
    def __init__(self, skip_traces: bool = True):
        from ..tools import eval

        self.eval = eval
        self.skip_traces = skip_traces

    def visit_Program(self, node: Program) -> Program:
        return Program(version=node.version, term=self.visit(node.term))

    def generic_visit(self, node: AST) -> AST:
        if self.skip_traces and TraceFinder().contains_trace(node):
            return super().generic_visit(node)
        try:
            nc = self.eval(node).result
        except Exception as e:
            nc = node
        else:
            if isinstance(nc, Exception) or not any(
                isinstance(nc, cfr) for cfr in _CONSTANT_FINAL_RESULTS
            ):
                nc = node
        return super().generic_visit(nc)
